#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <getopt.h>
#include <termios.h>
#include <string.h>

#define CMD_STATE	0xCC

struct reply_state {
    unsigned char cmd;          // [DM CMD]     0xCC             : 0 1 byte 
    unsigned char len;          // [PKT LENGTH] Length of packet : 1 1 byte 

/* High Data Rate mode */
    unsigned char hdr_mode;     // [HDR MODE]                    : 2 1 byte 
#define STATE_HDRM_AUTO		0x00    // Automatic 
#define STATE_HDRM_CDMA		0x02    // CDMA Only 
#define STATE_HDRM_HDR		0x02    // HDR Only 
#define STATE_HDRM_GPS		0x03    // GPS Only 

    unsigned char sys_srv_mode; // [SYS SRV MODE] System service mode : 3 1 byte 
#define STATE_SRVM_NO_SERVICE	0x00    // No Service 
#define STATE_SRVM_IS_95	0x01    // IS-95A or IS-95B 
#define STATE_SRVM_1x		0x02    // CDMA 1x 
#define STATE_SRVM_EVDO_R0	0x03    // CDMA EVDO in network Rev.0 
#define STATE_SRVM_EVDO_RA	0x04    // CDMA EVDO in network Rev.A 

    unsigned char rssi_1x;      // [1x RSSI] : 4 1 byte (dBm value) 
    unsigned char rssi_evdo;    // [EVDO RSSI] : 5 1 byte (dBm value) 
	//   If RSSI of value is 255(0xFF), means no service. 

    unsigned char ant_1x;       // [1x or EVDO ANT LEVEL] : 6 0-7

    unsigned char roam;	        // [ROAMING INDICATOR] : 7 1 byte 
#define STATE_ROAM_NONE		0x00    // Not Roaming 
#define STATE_ROAM_PREF		0x01    // Roaming inside preferred roaming list 
#define STATE_ROAM_NONPREF	0x02    // Roaming outside preferred roaming list  

    unsigned char pkt_state;    // [PKT STATE] : 8 1 byte 
#define STATE_STATE_IDLE	0x00	// Idle 
#define STATE_STATE_ACTIVE	0x01    // Active 
#define STATE_STATE_DORMANT	0x02    // Dormant 

    unsigned char err_code;     // [MIP ERR CODE] : 9 1 byte Mobile IP error code 
    unsigned long ltm_off;      // [LTM_OFF] : 10 Offset of local time from System Time.: 4 byte 
};

void usage(void)
{
    printf("Usage: modem_state [-1] [-v] [-h]\n");
    printf("This is tool for getting state of Cmotech CNx-550 modems\n");
    printf("  -1           single-shot run\n");
    printf("  -v           verbose output, shows bit state\n");
    printf("  -h           shows this help\n");
}

int main(int argc, char **argv)
{
  struct termios ts, ts_old;
  int ret, fd, verbose = 0;
  unsigned char cmd_state[] = {CMD_STATE, 0x18, 0xFC, 0x7E};
  unsigned char reply[256];
  struct reply_state *rs;
  char *device = "/dev/ttyUSB0";

    if(argc < 2) { usage(); return -1; }
  
    while ((ret = getopt(argc, argv, "1vh")) != -1)
    {
	switch(ret)
	{
	    case '1':
		break;
	    case 'v': 
		verbose = 1;
		break;
	    case 'h':
		usage();
		return 0;
	    default:
		usage();
		return -1;
	}
    }

    if(verbose) fprintf(stderr, "Opening %s\n", device);

    if((fd = open(device, O_RDWR)) >= 0)
    {


	tcdrain(fd);
	tcflush(fd, TCIOFLUSH);

	if(tcgetattr(fd, &ts) < 0)
	{
	    fprintf(stderr, "%s error: %s\n", device, strerror(errno));
	    close(fd);
	    return -1;
	}

	ts_old = ts;

	ts.c_cc[VMIN] = 0;
	ts.c_cc[VTIME] = 0;
	ts.c_oflag = 0;
	ts.c_lflag = 0;

	ts.c_cflag &= ~(CSIZE | CSTOPB | PARENB | CRTSCTS);
	ts.c_cflag |= CS8 | CREAD | CLOCAL;

	ts.c_iflag &= ~(IXON | IXOFF);
	ts.c_iflag |= IGNBRK | IGNPAR;

	ts.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ECHOKE);

	tcdrain(fd);

	if(verbose) fprintf(stderr, "Setting %s attrs\n", device);

	if(tcsetattr(fd, TCSANOW, &ts) < 0)
	{
	    fprintf(stderr, "%s error: %s\n", device, strerror(errno));
	    close(fd);
	    return -1;
	}

	if(verbose) fprintf(stderr, "Sending command to %s\n", device);

	if(write(fd, cmd_state, sizeof(cmd_state)) == sizeof(cmd_state))
	{
	    ret = read(fd, reply, sizeof(reply));

	    if(verbose) fprintf(stderr, "Received %d bytes from %s", ret, device);

	    switch(reply[0])
	    {
		case CMD_STATE:
		    rs = (struct reply_state *)reply;
		    printf("RSSI: 1x %d, EVDO %d", rs->rssi_1x, rs->rssi_evdo);
		    break;
	    }
	}

	if(verbose) fprintf(stderr, "Restoring %s attrs\n", device);

	if(tcsetattr(fd, TCSANOW, &ts_old) < 0)
	{
	    fprintf(stderr, "%s error: %s\n", device, strerror(errno));
	    close(fd);
	    return -1;
	}
	close(fd);
    }
    else
    {
	fprintf(stderr, "%s error: %s\n", device, strerror(errno));
	return -1;
    }

    return 0;
}
