/*
 * nvram - utility to work with nvram settings of Cmotech CNU-XXXpro
 *
 * Copyright (C) 2009 Andrew Kozin <stanson@pangolin.ru>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * mtd writing code borrowed from mtd utility from openwrt.org
 *
 *   mtd - simple memory technology device manipulation tool
 *
 *   Copyright (C) 2005 Waldemar Brodkorb <wbx@dass-it.de>,
 *	                    Felix Fietkau <nbd@openwrt.org>
 *
 */

#define _XOPEN_SOURCE
#define _GNU_SOURCE

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <getopt.h>
#include <unistd.h>
#include <linux/mtd/mtd.h>
#include <sys/ioctl.h>

#include "crc32.h"

#define CONFIG_SIZE	0x8000
#define CONFIG_OFFSET	0x8000
#define	MTD_DEV		"/dev/mtd1"
#define CONFIG_FILE	"/var/config.cache"
#define BUFSIZE		(8 * 1024)

struct config_header {
	unsigned long	tag; /* "FLSH" chars */
	unsigned long	size; /* size of data */
	unsigned long	crc_h; /* checksum of header */
	unsigned long 	crc_d; /* checksum of data */
};

struct config_header *config = NULL;
int old_size = 0, verbose = 0;
char buf[BUFSIZE];

void setup_header(struct config_header * hdr, int size)
{
	hdr->tag = 0x48534C46; /* FLSH */
	hdr->size = size;
	hdr->crc_d = ~crc32(0xFFFFFFFF, hdr + 1, hdr->size - sizeof(struct config_header));
	hdr->crc_h = 0;
	hdr->crc_h = ~crc32(0xFFFFFFFF, hdr, sizeof(struct config_header));
}

int check_header(struct config_header * hdr)
{
	struct config_header h;
	unsigned long	crc;

	h.tag = hdr->tag;
	h.size = hdr->size;
	h.crc_h = 0;
	h.crc_d = hdr->crc_d;

	if(hdr->tag != 0x48534C46)
	{
	    fprintf(stderr, "Bad header tag\n");
	    return 0;
	}

	if(hdr->size & 3 || hdr->size > CONFIG_SIZE)
	{
	    fprintf(stderr, "Bad size or alignment. size=%lu\n", hdr->size);
	    return 0;
	}

	crc = ~crc32(0xFFFFFFFF, &h, sizeof(h));
	if(hdr->crc_h != crc)
	{
	    fprintf(stderr, "Bad header checksum. Header: 0x%08lX CRC32: 0x%08lX\n", 
				hdr->crc_h, crc);
	    return 0;
	}

	crc = ~crc32(0xFFFFFFFF, hdr + 1, hdr->size - sizeof(struct config_header));
	if(hdr->crc_d != crc)
	{
	    fprintf(stderr, "Bad data checksum. Header: 0x%08lX CRC32: 0x%08lX\n", 
				hdr->crc_d, crc);
	    return 0;
	}
	return 1;	
}

struct config_header * open_config(const char *fn)
{
	int fd, mtd, s;
	struct config_header *conf = NULL;

	fd = open(fn, O_RDWR);
	if(fd < 0)
	{
		if(errno == ENOENT)
		{
			fd = open(fn, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
			if(fd < 0)
			{
				fprintf(stderr, "%s error: %s\n", fn, strerror(errno));
				return NULL;
			}

			mtd = open(MTD_DEV, O_RDONLY);
			if(mtd < 0)
			{
				fprintf(stderr, "%s error: %s\n", MTD_DEV, strerror(errno));
				goto err_fd;
			}
			
			/* copy mtd contents to cache file */
			while((s = read(mtd, buf, sizeof(buf))) > 0)
			{
				if(write(fd, buf, s) < 0)
				{
					fprintf(stderr, "%s write error: %s\n", fn, strerror(errno));
					goto err_mtd;
				}
			}
			if(s < 0)
			{
				fprintf(stderr, "%s read error: %s\n", MTD_DEV, strerror(errno));
				goto err_mtd;
			}
		}
		else
		{
			fprintf(stderr, "%s error: %s\n", fn, strerror(errno));
			return NULL;
		}
	}

	if(fd >= 0)
	{
		lseek(fd, CONFIG_OFFSET, SEEK_SET);
		conf = malloc(CONFIG_SIZE);
		if(conf)
		{
			s = read(fd, conf, CONFIG_SIZE);
			if(s < 0)
			{
				fprintf(stderr, "%s read error: %s\n", fn, strerror(errno));
				goto err_fd;
			}
			if(s != CONFIG_SIZE)
			{
				fprintf(stderr, "%s read error: only %d bytes read\n", fn, s);
				goto err_fd;
			}
		}
	}
	close(fd);
	return conf;

err_mtd:
	close(mtd);
err_fd:
	close(fd);
	if(conf) free(conf);
	return NULL;
}

int write_config(char *fn, struct config_header *conf)
{
	int fd, ret;

	fd = open(fn, O_RDWR);
	if(fd < 0)
	{
		fprintf(stderr, "%s write error: %s\n", fn, strerror(errno));
		return -1;
	}

	lseek(fd, CONFIG_OFFSET, SEEK_SET);

	ret = write(fd, conf, CONFIG_SIZE);
	if(ret < 0)
	{
		fprintf(stderr, "%s write error: %s\n", fn, strerror(errno));
		goto err_out;
	}
	if(ret != CONFIG_SIZE)
	{
		fprintf(stderr, "%s write error: only %d bytes written\n", fn, ret);
		ret = -1;
		goto err_out;
	}
	fsync(fd);
err_out:
	close(fd);
	return ret;
}

int commit_config(char *fn)
{
	int fd, imagefd, result, ret = -1;
	size_t r, w, e;
	struct mtd_info_user mtdInfo;
	struct erase_info_user mtdEraseInfo;

	fd = open(MTD_DEV, O_RDWR | O_SYNC);
	if(fd < 0) 
	{
		fprintf(stderr, "Could not open mtd device: %s\n", MTD_DEV);
		goto err_nomtd;
	}

	if(ioctl(fd, MEMGETINFO, &mtdInfo)) 
	{
		fprintf(stderr, "Could not get MTD device info from %s\n", MTD_DEV);
		goto err_noin;
	}

	if(verbose)
	{
		fprintf(stderr, "%s MTD info:\n", MTD_DEV);
		fprintf(stderr, "  Type:       0x%02X\n", mtdInfo.type);
		fprintf(stderr, "  Flags:      0x%08X\n", mtdInfo.flags);
		fprintf(stderr, "  Size:       0x%08X\n", mtdInfo.size);
		fprintf(stderr, "  Erase size: 0x%08X\n", mtdInfo.erasesize);
		fprintf(stderr, "  OOB blocks: 0x%08X\n", mtdInfo.oobblock);
		fprintf(stderr, "  OOB size:   0x%08X\n", mtdInfo.oobsize);
		fprintf(stderr, "  ECC type:   0x%08X\n", mtdInfo.ecctype);
		fprintf(stderr, "  ECC size:   0x%08X\n", mtdInfo.eccsize);
	}

	imagefd = open(fn, O_RDWR);

	if(imagefd < 0)
	{
		fprintf(stderr, "%s error: %s\n", fn, strerror(errno));
		goto err_noin;
	}

	r = w = e = 0;

	if(verbose) fprintf(stderr, "Writing data to %s ", MTD_DEV);

	for (;;) {
		r = read(imagefd, buf, sizeof(buf));
		w += r;

		/* EOF */
		if (r <= 0) break;

		/* need to erase the next block before writing data to it */
		while (w > e) {
			mtdEraseInfo.start = e;
			mtdEraseInfo.length = mtdInfo.erasesize;
			if(verbose) fprintf(stderr, "-");
			/* erase the chunk */
			if (ioctl(fd,MEMERASE,&mtdEraseInfo) < 0) {
				fprintf(stderr, "Erasing mtd failed: %s\n", MTD_DEV);
				goto err;
			}
			e += mtdInfo.erasesize;
		}
		
		if(verbose) fprintf(stderr, "+");

		result = write(fd, buf, r);
		if (result < 0) {
			fprintf(stderr, "Error writing image.\n");
			goto err;
		} else {
			if ((unsigned)result < r) {
				fprintf(stderr, "Insufficient space.\n");
				goto err;
			}
		}
	}

	if(verbose) fprintf(stderr, " Done\n");

	ret = 0;
err:
	close(imagefd);
err_noin:
	close(fd);
err_nomtd:
	return ret;
}

void dump_config(struct config_header *conf)
{
	char	*s = (char *)(conf + 1);

	if(conf)
	{
		while(*s)
		{
			fprintf(stdout, "%s\n", s);
			s += strlen(s) + 1;
		}
	}
}

int set_val(struct config_header *conf, const char *name, int del)
{
	const char *s1;
	char *s2, *s;
	int nl, ol, size, found;

	if(!conf || !name) return 0;

	if(!del && !strchr(name, '=')) return 0;

	nl = strlen(name) + 1;
	size = sizeof(struct config_header);
	found = 0;

	s = (char *)(conf + 1);

	while(*s)
	{
		s1 = name;
		s2 = s;
		ol = strlen(s) + 1;

		if(!found)
		{
			while(*s1 && *s1 != '=' && *s2 && *s2 != '=')
			{
				if(*s2 != *s1) break;
				s1++; s2++;
			}

			if(!del && *s1 == '=' && *s2 == '=')
			{
				/* found name, replace name=val with new data */
				if(conf->size + nl - ol > CONFIG_SIZE) goto err_large;
				memmove(s + nl, s + ol, conf->size - (size + ol));
				strcpy(s, name);
				ol = nl;
				found = 1;
			}
			
			if(del && *s1 == 0 && *s2 == '=')
			{
				memmove(s, s + ol, conf->size - (size + ol));
				ol = 0;
				found = 1;
			}
		}
		
		s += ol;
		size += ol;
	}

	if(!found && del)
	{
		fprintf(stderr, "Name %s not found!\n", name);
		return 0;
	}
	
	if(!found)
	{
		/* name not found, just add name=val pair to the end */
		if(conf->size + nl > CONFIG_SIZE) goto err_large;
		strcpy(s, name);
		s += nl;
		size += nl;

	}

	nl = ((size + 5) & ~3) - size;
	memset(s, 0, nl);
	s += nl;
	size += nl;
	memset(s, 0xFF, CONFIG_SIZE - size);
	setup_header(conf, size);
	return 1;

err_large:
	fprintf(stderr, "Size of new data is too large: %d\n", nl);
	return 0;

}

char * get_val(struct config_header *conf, const char *name)
{
	char	*s, *s2, *val, tmp[16], *pass;
	const char *s1;
	unsigned long num;

	if(!conf || !name) return NULL;

	s = (char *)(conf + 1);
	while(*s)
	{
		s1 = name;
		s2 = s;

		while(*s1 && *s2 && *s2 != '=')
		{
			if(*s2 != *s1) break;
			s1++; s2++;
		}

		if(*s1 == 0 && *s2 == '=')
		{
			/* found name, return val*/
			val = strdup(s2 + 1);
			if(val)
			{
				if(val[0] == '0' && val[1] == 'x')
				{
					num = strtoul(val, NULL, 16);
					free(val);
					snprintf(tmp, sizeof(tmp), "%lu", num);
					val = strdup(tmp);
				}
				if(!strcmp(name,"sysmode0/adminpass"))
				{
					pass = crypt(val,"$1$");
					if(pass)
					{
						free(val);
						val = strdup(pass);
					}
				}
				return val;
			}
		}
		
		s += strlen(s) + 1;
	}
	return NULL;
}

int xlate_file(struct config_header *conf, char * fn)
{
	int		ret = 0, i = 0;
	FILE		*f;
	char		*p, *start, *end, *e, *val;

	f = fopen(fn, "r");
	if(!f)
	{
		fprintf(stderr, "%s error: %s\n", fn, strerror(errno));
		ret = 1;
		goto exit_noin;
	}

	while(fgets(buf, sizeof(buf), f))
	{
		p = buf;
		while((start = strstr(p, "<%")))
		{
			if((end = strstr(start, "%>")))
			{
				*start = 0; start += 2;
				printf("%s", p);

				/* skip leading spaces */
				while(*start && *start == ' ') start++;
				/* strip trailing spaces */
				for(e = start; *e && *e != ' '; e++); *e = 0;
				val = get_val(conf, start);
				if(val)
				{
					printf("%s", val);
					free(val);
				}

				p = end + 2;
			}
			else
			{
				fprintf(stderr,"Parse error at %s:%d, no closing '%%>'\n", fn, i);
				ret = 1;
				break;
			}
		}
		if(*p) printf("%s", p);
	}

	fclose(f);
exit_noin:	
	return ret;
}

void usage(void)
{
    printf("Usage: nvram ([-x file] | [-g name] | [-s name=val] | [-d name]) [-l] [-c] [-v] [-h]\n");
    printf("This is a tool for working with nvram settings in CNU-XXXpro.\n");
    printf("Licensed under GPLv3. Copyright (C) 2009 Andrew Kozin.\n");
    printf("  -x <file>       translate <file> using nvram settings to stdout\n");
    printf("  -s <name>=<val> set value for <name> in nvram\n");
    printf("  -g <name>       get value for <name> from nvram\n");
    printf("  -d <name>       delete <name> from nvram\n");
    printf("  -l              list all settings\n");
    printf("  -c              commit changes (actually write to flash)\n");
    printf("  -v              verbose output\n");
    printf("  -h              shows this help\n");
}

#define	CMD_NONE	0
#define CMD_XLATE	1
#define CMD_SET		2
#define CMD_GET		3
#define CMD_DELETE	4
#define CMD_COMMIT	5
#define CMD_LIST	6

int main(int argc, char **argv)
{
    int ret, cmd = CMD_NONE;
    char *file = NULL;
    char *name = NULL;
    char *val = NULL;

    if(argc < 2) { usage(); return -1; }
  
    while ((ret = getopt(argc, argv, "x:g:s:d:lcvh")) != -1)
    {
	switch(ret)
	{
	    case 'x':
		cmd = CMD_XLATE; file = strdup(optarg);
		break;
	    case 'g':
		cmd = CMD_GET; name = strdup(optarg);
		break;
	    case 's':
		cmd = CMD_SET; name = strdup(optarg);
		break;
	    case 'd':
		cmd = CMD_DELETE; name = strdup(optarg);
		break;
	    case 'l':
		cmd = CMD_LIST;
		break;
	    case 'c':
		cmd = CMD_COMMIT;
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'h':
		usage();
		break;
	    default:
		usage();
		return -1;
	}
    }

    if(cmd != CMD_NONE)
    {
	config = open_config(CONFIG_FILE);
	if(config)
	{
	    check_header(config);
	    switch(cmd)
	    {
		case CMD_LIST:
		    dump_config(config);
		    break;
		case CMD_XLATE:
		    xlate_file(config, file);
		    break;
		case CMD_GET:
		    val = get_val(config, name);
		    if(val)
			printf("%s\n", get_val(config, name));
		    break;
		case CMD_SET:
		    if(set_val(config, name, 0))
			write_config(CONFIG_FILE, config);
		    break;
		case CMD_DELETE:
		    if(set_val(config, name, 1))
			write_config(CONFIG_FILE, config);
		case CMD_COMMIT:
		    commit_config(CONFIG_FILE);
		    break;
	    }
	    check_header(config);
	    free(config);
	}
    }

    if(file) free(file);
    if(name) free(name);

    return 0;
}
